from __future__ import annotations

import logging
from typing import Dict, Optional, Any
from datetime import datetime as dt

import requests
from faker import Faker
from sqlalchemy import func

import ckan.plugins.toolkit as tk
import ckan.logic as logic
import ckan.lib.mailer as mailer
from ckan.lib.munge import munge_name
from ckan.common import session
from ckan.model import User, Session

import ckanext.msal.config as conf
import ckanext.msal.utils as msal_utils


log = logging.getLogger(__name__)

user_show = tk.get_action("user_show")
user_update = tk.get_action("user_update")

USER_ENDPOINT = "https://graph.microsoft.com/beta/me"


def _login_user(user_data: dict) -> Dict[str, Any]:
    return get_or_create_user(user_data)


def get_or_create_user(user_data: Dict[str, Any]) -> Dict[str, Any]:
    """Returns an existed user data by oid (object id) or
    creates a new one from user_data fetched from Microsoft Graph API

    args
    user_data: dict - user data dict

    {
        "displayName": "Mark Spencer",
        "givenName": "Mark",
        "mailNickname": "mark209",
        "mail": mark209@myorg.onmicrosoft.com,
        "surname": "Spencer",
        "userPrincipalName": "mark209@myorg.onmicrosoft.com",
        "id": "3f22cb88-c272-44f1-838f-f823cdc08bd6",
    }

    return
    type: Dict[str, Any]
    """

    try:
        user: Dict[str, Any] = _get_user(user_data)
    except tk.ObjectNotFound:
        log.info(f"MSAL. User not found, creating new one.")
        return _create_user_from_user_data(user_data)

    return user


def _get_user(user_data: dict[str, Any]) -> dict[str, Any]:
    """Searches for an existing user created with MSAL
    or for a user with the same email


    Args:
        user_data (dict[str, Any]): MSAL user data

    Raises:
        tk.ObjectNotFound: raises an error if there is no user

    Returns:
        dict[str, Any]: CKAN user data
    """

    user = (
        Session.query(User.id)
        .filter(User.plugin_extras["msal"]["id"].astext == str(user_data["id"]))
        .one_or_none()
    )

    if not user and tk.config.get(
        conf.MERGE_MATCHING_EMAILS, conf.MERGE_MATCHING_EMAILS_DF
    ):
        user_dict = _merge_users(user_data)

        if user_dict:
            return user_dict

    if not user:
        raise tk.ObjectNotFound(
            tk._(f"User with MSAL ID - {user_data['id']} not found")
        )

    return user_show(msal_utils.get_site_admin_context(), {"id": user.id})


def _merge_users(user_data: dict) -> Optional[dict[str, Any]]:
    user_email: str = _get_email(user_data)
    query = Session.query(User)
    _lower = func.lower if is_email_insensitive() else lambda x: x
    query = query.filter(_lower(User.email) == _lower(user_email))

    user = query.one_or_none()

    if user is None:
        return

    log.info(f"MSAL. A user with the same email has been found: {user_email}")
    log.info("MSAL. Merging users.")

    context = msal_utils.get_site_admin_context()

    user_dict = user_show(context, {"id": user.id})
    user_dict.setdefault("plugin_extras", {})

    user_dict["plugin_extras"]["msal"] = {"id": user_data["id"]}
    user_dict["password"] = msal_utils._make_password()
    user_dict["email"] = user_email

    user_obj = context["user_obj"]
    try:
        log.info(f"Emailing reset link to user: {user_obj.name}")
        mailer.send_reset_link(user_obj)
    except mailer.MailerException as e:
        # SMTP is not configured correctly or the server is
        # temporarily unavailable
        log.error(tk._("MSAL. Error sending the password reset email."))
        log.error(e)

    return user_update(context, user_dict)


def get_msal_user_data() -> Dict[str, Any]:
    """
    Requests an additional user data from microsoft graph API

    return
    type: Dict[str, Any]
    """
    token: Optional[Dict[Any, Any]] = msal_utils._get_token_from_cache(conf.SCOPE)

    access_token: str = session.get("msal_auth_flow", {}).get("access_token", "")
    if not access_token:
        token: Optional[Dict[Any, Any]] = msal_utils._get_token_from_cache(conf.SCOPE)
        access_token: str = token["access_token"] if token else ""

        if not access_token:
            return {"error": [tk._("The token has expired. Please, try again.")]}

    log.info(f"MSAL. Fetching user data from Microsoft Graph API by an access token.")
    resp = requests.get(
        USER_ENDPOINT,
        headers={"Authorization": "Bearer " + access_token},
        params={
            "$id"
            "$select": "id,displayName,userPrincipalName,mail,mailNickname,accountEnabled"
        },
    )

    try:
        resp.raise_for_status()
    except requests.HTTPError:
        log.error(f"MSAL. Fetch failed: {resp.reason} {resp.status_code}")
        return resp.json()

    log.info(f"MSAL. Success fetch.")

    user_data: dict[str, Any] = resp.json()
    user_email: str = _get_email(user_data)

    if msal_utils.is_email_restricted(user_email) or not msal_utils.is_email_allowed(
        user_email
    ):
        log.info(
            "MSAL. User won't be created, "
            f"because of the domain policy: {user_email}"
        )

        restriction_err: str = tk.config.get(
            conf.RESTRICTION_ERR, conf.RESTRICTION_ERR_DF
        )
        raise tk.ValidationError({"email": [tk._(restriction_err)]})
    return resp.json()


def _create_user_from_user_data(user_data: dict) -> Dict[str, Any]:
    """Create a user with random password using Microsoft Graph API's data.

    raises
    ValidationError if email is not unique

    args
    user_data: dict - user data dict
    object_id: str - actually a `user_id` from Azure AD

    return
    type: dict
    """

    email: str = _get_email(user_data)
    password: str = msal_utils._make_password()
    username: str = munge_name(_get_username(user_data))

    if not _is_username_unique(username):
        username = f"{username}-{dt.now().strftime('%S%f')}"

    user = tk.get_action("user_create")(
        msal_utils.get_site_admin_context(),
        {
            "email": email,
            "name": username,
            "password": password,
            "plugin_extras": {"msal": {"id": user_data["id"]}},
        },
    )
    log.info(f"MSAL. User has been created: {user['id']} - {user['name']}.")
    return user


def _get_email(
    user_dict: Dict[
        str,
        str,
    ]
) -> str:
    """
    Fetches email from user_data if exists, otherwise generates random email
    The `userPrincipalName` is formatted like an email address (username@onmicrosoft.com)

    userPrincipalName: The user principal name (UPN) of the user.
        The UPN is an Internet-style login name for the user based on
        the Internet standard RFC 822. By convention,
        this should map to the user's email name.

    mail: The SMTP address for the user, for example, 'jeff@contoso.onmicrosoft.com'
    """
    return user_dict.get("userPrincipalName") or user_dict.get("mail") or _make_email()


def _make_email(domain: str = "msal.onmicrosoft.com") -> str:
    """
    Returns a random email with custom domain
    If domain is not provided uses `onmicrosoft.com`

    args
    domain: str - domain used to generate email

    return
    type: str
    """
    f = Faker()
    return f.email(domain)


def _get_username(user_dict: Dict[str, str]) -> str:
    """
    Fetches username from user_data if exists
    If not - munges it from userPrincipalName or mail

    args
    user_dict: Dict[str, str] - user data dict

    return
    type: str
    """
    username: Optional[str] = user_dict.get("mailNickname")

    if not username:
        username = user_dict.get("mail") or user_dict["userPrincipalName"]
        return username.split("@")[0]
    return username


def _is_username_unique(username: str) -> bool:
    try:
        user_show({"ignore_auth": True}, {"id": username})
    except logic.NotFound:
        return True

    return False


def is_user_enabled(user_dict: Dict[str, str]) -> bool:
    """
    Returns True if user is enabled

    args
    user_dict: Dict[str, str] - user data dict

    return
    type: bool
    """
    return user_dict.get("accountEnabled")


def is_user_sysadmin(user_dict: Dict[str, str]) -> bool:
    """
    Returns True if user is sysadmin

    # TODO
    Currently, we are not using this function.
    I didn't find an apropriate property to say if user is a sysadmin yet

    args
    user_dict: Dict[str, str] - user data dict

    return
    type: bool
    """
    return False


def is_email_insensitive() -> bool:
    return tk.asbool(
        tk.config.get(conf.EMAIL_CASE_INSENSITIVE, conf.EMAIL_CASE_INSENSITIVE_DF)
    )



import logging
from datetime import datetime as dt

from flask import Blueprint

import ckan.lib.helpers as h
import ckan.plugins.toolkit as tk
from ckan.common import session

import ckanext.msal.config as conf
import ckanext.msal.utils as msal_utils
from ckanext.msal.user import get_msal_user_data


log = logging.getLogger(__name__)
msal = Blueprint("msal", __name__)


@msal.route(tk.config.get(conf.REDIRECT_PATH, conf.REDIRECT_PATH_DF))
def authorized():
    try:
        cache = msal_utils._load_cache()
        result = msal_utils.build_msal_app(cache=cache).acquire_token_by_auth_code_flow(
            session.get("msal_auth_flow", {}), tk.request.args
        )

        session["msal_auth_flow"] = result

        try:
            user_data = get_msal_user_data()
        except tk.ValidationError as e:
            msal_utils._flash_validation_errors(e)
            return h.redirect_to(h.url_for("user.login"))

        if "error" in result or "error" in user_data:
            msal_utils._clear_session()
            log.error(result["error"] or user_data["error"])
            h.flash_error(tk._("Login error. Contact administrator."))
            return h.redirect_to(h.url_for("user.login"))
    
        session["user"] = user_data
        session["user_exp"] = msal_utils._get_exp_date()
        msal_utils._save_cache(cache)
    except ValueError:
        # Usually caused by CSRF
        # Simply ignore them
        pass

    return h.redirect_to(h.url_for("dashboard.index"))


@msal.route("/user/msal-logout")
def logout():
    if session.get("msal_auth_flow") or session.get("msal_token_cache"):
        msal_utils._clear_session()
        redirect_uri: str = h.url_for("user.logout", _external=True)
        authority: str = tk.config.get(conf.AUTHORITY, conf.AUTHORITY_DF)
        authority_url: str = f"https://login.microsoftonline.com/{authority}"
        return h.redirect_to(
            f"{authority_url}/oauth2/v2.0/logout?post_logout_redirect_uri={redirect_uri}"
        )

    return h.redirect_to("user.logout")


@msal.route("/user/msal-login")
def login():
    flow = msal_utils.build_auth_code_flow(scopes=conf.SCOPE)
    session["msal_auth_flow"] = flow

    return h.redirect_to(flow["auth_uri"], _external=True)


def get_blueprints():
    return [msal]



from datetime import datetime as dt

import pytest

import ckanext.msal.utils as utils


@pytest.mark.ckan_config("ckan.plugins", "msal")
@pytest.mark.usefixtures("with_plugins")
class TestUtils(object):
    user_dict = {
        "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users/$entity",
        "@odata.id": "https://graph.microsoft.com/v2/3c8827a9-65fe-40b5-8644-3173d7026601/directoryObjects/fb9c93ba-0768-4816-8fcc-802b588fb8bf/Microsoft.DirectoryServices.User",
        "businessPhones": ["380666333211"],
        "displayName": "Mark Spencer",
        "givenName": "Mark",
        "mailNickname": "mark209",
        "jobTitle": None,
        "mail": None,
        "mobilePhone": None,
        "officeLocation": None,
        "preferredLanguage": None,
        "surname": "Spencer",
        "userPrincipalName": "kvaqich@kvaqich.onmicrosoft.com",
        "id": "fb9c93ba-0768-4816-8fcc-802b588fb8bf",
    }

    def test_exp_date(self):
        exp_date: float = utils._get_exp_date()

        assert isinstance(exp_date, float)
        assert dt.now().timestamp() < exp_date

    def test_make_password(self):
        pass1 = utils._make_password()
        pass2 = utils._make_password()

        assert len(pass1) > 16
        assert pass1 != pass2



import ckan.plugins as p
import ckan.plugins.toolkit as tk
import ckan.model as model
from ckan.common import session

from ckanext.msal.middleware import SessionInvalidator
from ckanext.msal.views import get_blueprints
import ckanext.msal.user as user_funcs
import ckanext.msal.utils as msal_utils


class MsalPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurer)
    p.implements(p.IAuthenticator, inherit=True)
    p.implements(p.IMiddleware, inherit=True)
    p.implements(p.IBlueprint, inherit=True)

    # IConfigurer

    def update_config(self, config_):
        tk.add_template_directory(config_, "templates")

    # IMiddleware
    def make_middleware(self, app, config):
        return SessionInvalidator(app)

    # IBlueprint
    def get_blueprint(self):
        return get_blueprints()

    # IAuthenticator
    def identify(self):
        """Called to identify the user.

        If the user is identified then it should set:

         - tk.g.user: The name of the user
         - tk.g.userobj: The actual user object
        """

        if session.get("user") and not any(
            (tk.g.setdefault("userobj"), tk.g.setdefault("user"))
        ):
            
            try:
                user = user_funcs._login_user(session["user"])
            except tk.ValidationError as e:
                return msal_utils._flash_validation_errors(e)

            if user:
                tk.g.user = user["name"]
                tk.g.userobj = model.User.get(user["id"])
            else:
                msal_utils._clear_session()

    def logout(self):
        msal_utils._clear_session()



import ckan.plugins.toolkit as tk
from ckan.common import config


CLIENT_ID = "ckanext.msal.client_id"
CLIENT_SECRET = "ckanext.msal.client_secret"
AUTHORITY = "ckanext.msal.tenant_id"
AUTHORITY_DF = "common"
REDIRECT_PATH = "ckanext.msal.redirect_path"
REDIRECT_PATH_DF = "/get_msal_token"
USER_SESSION_TTL = "ckanext.msal.session_lifetime"
USER_SESSION_TTL_DF = 3600
RESTRICTED_DOMAINS = "ckanext.msal.restrict.restricted_domain_list"
ALLOWED_DOMAINS = "ckanext.msal.restrict.allowed_domain_list"
RESTRICTION_ERR = "ckanext.msal.restrict.error_message"
RESTRICTION_ERR_DF = "Your email domain is restricted. Please, contact site admin."

EMAIL_CASE_INSENSITIVE = "ckanext.msal.email_case_insensitive"
EMAIL_CASE_INSENSITIVE_DF = False

MERGE_MATCHING_EMAILS = "ckanext.msal.merge_matching_emails"
MERGE_MATCHING_EMAILS_DF = False

SCOPE = ["User.ReadBasic.All"]



from datetime import datetime as dt

from ckan.common import session

import ckanext.msal.utils as msal_utils
from ckanext.msal.user import get_msal_user_data, is_user_enabled


def _invalidate_user_session():
    """
    Clears session if user session is expired
    Otherwise, sets a new expiration date
    """

    if session.get("user"):
        if _is_session_expired():
            if not is_logged_in():
                msal_utils._clear_session()
            else:
                session["user_exp"] = msal_utils._get_exp_date()


def _is_session_expired() -> bool:
    """
    Returns `True` if user session is expired

    return
    type: bool
    """

    if dt.now().timestamp() >= session.get("user_exp", 0):
        return True
    return False


def is_logged_in():
    user_data = get_msal_user_data()
    # TODO: hypothetically, here we can change something in user metadata

    if "error" in user_data or not is_user_enabled(user_data):
        return False

    return True


class SessionInvalidator(object):
    def __init__(self, app):
        """Initialize the Session Invalidator Middleware"""
        self.app = app
        app.before_request(_invalidate_user_session)

    def __call__(self, environ, start_response):

        return self.app(environ, start_response)



import secrets
from typing import Dict, Optional, Any, List
from datetime import datetime as dt
from datetime import timedelta as td

import msal

import ckan.lib.helpers as h
import ckan.plugins.toolkit as tk
from ckan.common import session

import ckanext.msal.config as conf


def build_msal_app(cache=None):
    authority: str = tk.config.get(conf.AUTHORITY, conf.AUTHORITY_DF)
    client_id: str = tk.config.get(conf.CLIENT_ID)
    client_credential: str = tk.config.get(conf.CLIENT_SECRET)

    return msal.ConfidentialClientApplication(
        client_id,
        authority=f"https://login.microsoftonline.com/{authority}",
        client_credential=client_credential,
        token_cache=cache,
    )


def build_auth_code_flow(authority=None, scopes=None) -> Dict[str, Any]:
    return build_msal_app().initiate_auth_code_flow(
        scopes or [], redirect_uri=h.url_for("msal.authorized", _external=True)
    )


def _load_cache() -> msal.SerializableTokenCache:
    cache = msal.SerializableTokenCache()
    if session.get("msal_token_cache"):
        cache.deserialize(session["msal_token_cache"])
    return cache


def _save_cache(cache) -> None:
    if cache.has_state_changed:
        session["msal_token_cache"] = cache.serialize()


def _get_token_from_cache(scope=None) -> Optional[Dict[Any, Any]]:
    cache = _load_cache()
    app = build_msal_app(cache=cache)
    accounts: List[Dict[Any, Any]] = app.get_accounts()

    if accounts:
        _save_cache(cache)
        return app.acquire_token_silent(scope, account=accounts[0])
    return app.acquire_token_silent(scope, account=None)

def _get_exp_date():
    """
    Returns a float number that represents an expiration date of user session
    The session lifetime is configurable with `ckanext.msal.session_lifetime` option

    return
    type: float
    """

    session_ttl: int = tk.asint(
        tk.config.get(conf.USER_SESSION_TTL, conf.USER_SESSION_TTL_DF)
    )
    return (dt.now() + td(seconds=session_ttl)).timestamp()


def _make_password() -> str:
    """
    Return a random URL-safe text string, in Base64 encoding

    return
    type: str
    """
    return secrets.token_urlsafe(60)


def get_restricted_domains() -> List[str]:
    """Returns a lits of restricted domains from config
    User won't be able to login with this email domain

    Returns:
        List[str]: a list of domain strings
    """

    return tk.aslist(tk.config.get(conf.RESTRICTED_DOMAINS), ",")


def get_allowed_domains() -> List[str]:
    """Returns a lits of allowed domains from config
    User will be able to login only with those email domains

    Returns:
        List[str]: a list of domain strings
    """

    return tk.aslist(tk.config.get(conf.ALLOWED_DOMAINS), ",")


def is_email_restricted(email: str) -> bool:
    """Checks if the user email is restricted by domain
    Returns True if resticted

    Args:
        email (str): user email

    Returns:
        bool
    """

    for domain in get_restricted_domains():
        if email.endswith(domain):
            return True
    return False


def is_email_allowed(email: str) -> bool:
    """Checks if the user email is allowed by domain
    Returns True if allowed

    Args:
        email (str): user email

    Returns:
        bool
    """

    allowed_domains: list[str] = get_allowed_domains()

    if not allowed_domains:
        return True

    for domain in get_allowed_domains():
        if email.endswith(domain):
            return True
    return False


def get_site_admin_context() -> dict:
    site_user = tk.get_action("get_site_user")({"ignore_auth": True}, {})
    return {"user": site_user["name"], "ignore_auth": True}


def _clear_session():
    """Wipe out user and its token cache from session
    """
    session.clear()


def _flash_validation_errors(error: dict):
    """Adds validation errors to the session

    Args:
        error (dict): tk.ValidationError error_dict
    """
    _clear_session()
    session["flash"] = []

    for e in error.error_summary.items():
        session["flash"].append(
            ("alert-error", f"{e[1]}", True)
        )


from uuid import uuid4
from datetime import datetime as dt

import pytest

import ckan.logic as logic
from ckan.tests import factories

import ckanext.msal.user as user_funcs


@pytest.mark.ckan_config("ckan.plugins", "msal")
@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestUser(object):
    user_dict = {
        "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users/$entity",
        "@odata.id": "https://graph.microsoft.com/v2/3c8827a9-65fe-40b5-8644-3173d7026601/directoryObjects/fb9c93ba-0768-4816-8fcc-802b588fb8bf/Microsoft.DirectoryServices.User",
        "businessPhones": ["380999222611"],
        "displayName": "Mark Spencer",
        "givenName": "Mark",
        "mailNickname": "mark209",
        "jobTitle": None,
        "mail": None,
        "mobilePhone": None,
        "officeLocation": None,
        "preferredLanguage": None,
        "surname": "Spencer",
        "userPrincipalName": "kvaqich@kvaqich.onmicrosoft.com",
        "id": "fb9c93ba-0768-4816-8fcc-802b588fb8bf",
    }

    def test_get_email_upn(self):
        """
        Returns userPrincipalName from user_dict
        """
        email = user_funcs._get_email(self.user_dict)
        assert email == "kvaqich@kvaqich.onmicrosoft.com"

    def test_get_email_mail(self):
        """
        Returns `mail` field if `userPrincipalName` is empty
        """
        email = user_funcs._get_email({"mail": "kvaqich@kvaqich.onmicrosoft.com"})
        assert email == "kvaqich@kvaqich.onmicrosoft.com"

    def test_get_email_empty(self):
        """
        Returns random email if both fields are empty
        """
        email = user_funcs._get_email({})
        assert email
        assert email.endswith("msal.onmicrosoft.com")

    def test_get_username(self):
        """
        Fetches username from user_dict
        mailNickname -> munge(mail) -> munge(userPrincipalName)
        """
        username = user_funcs._get_username(self.user_dict)
        username == "mark209"

        username = user_funcs._get_username({"mail": "testuser@gmail.com"})
        username == "testuser"

        username = user_funcs._get_username(
            {"userPrincipalName": "testuser2@gmail.com"}
        )
        username == "testuser2"

    def test_is_username_unqiue(self):
        user = factories.User()

        is_unique: bool = user_funcs._is_username_unique(user["name"])
        assert not is_unique

        is_unique: bool = user_funcs._is_username_unique("absolutely_new_user")
        assert is_unique

    def test_create_and_get(self):
        user_data = {
            "userPrincipalName": "mark.spencer@linkdigital.com.au",
            "id": str(uuid4()),
        }
        user1 = user_funcs.get_or_create_user(user_data)

        assert user1["plugin_extras"]["msal"]["id"] == user_data["id"]
        assert user1["name"] == "mark-spencer"

        user2 = user_funcs.get_or_create_user(user_data)

        assert user1["id"] == user2["id"]

    def test_user_merging_disabled(self):
        factories.User(email="joe.black@linkdigital.com.au")

        user_data = {
            "userPrincipalName": "joe.black@linkdigital.com.au",
            "id": str(uuid4()),
        }

        with pytest.raises(logic.ValidationError):
            user_funcs.get_or_create_user(user_data)

    @pytest.mark.ckan_config("ckanext.msal.merge_matching_emails", "True")
    @pytest.mark.usefixtures("with_request_context")
    def test_user_merging_enabled(self):
        user1 = factories.User(email="joe.black@linkdigital.com.au")

        user_data = {
            "userPrincipalName": "joe.black@linkdigital.com.au",
            "id": str(uuid4()),
        }

        user2 = user_funcs.get_or_create_user(user_data)

        assert user1["id"] == user2["id"]
        assert user2["plugin_extras"]["msal"]["id"] == user_data["id"]
